package cn.ibizlab.core.extensions.listener;

import lombok.extern.slf4j.Slf4j;

import org.flowable.bpmn.model.UserTask;
import org.flowable.common.engine.api.delegate.event.*;
import org.flowable.common.engine.api.delegate.event.FlowableEngineEventType;
import org.flowable.common.engine.impl.event.FlowableEntityEventImpl;
import org.flowable.task.service.impl.persistence.entity.TaskEntity;
import org.flowable.task.service.impl.persistence.entity.TaskEntityImpl;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.util.ObjectUtils;

import java.util.List;

/**
 * 流程实例监听器，监听流程实例启动、流转、结束状态
 * 1.流程状态发生变化时，将实例信息同步到统一工作流平台
 * 2.到达流程节点时，调用统一接口，查询当前节点办理人并设置到引擎中
 */
@Slf4j
@Component
public class ProcessInstanceListener implements FlowableEventListener {

    @Override
    public void onEvent(FlowableEvent evt) {

        if(evt instanceof FlowableEntityEventImpl)
        {
            FlowableEntityEventImpl event=((FlowableEntityEventImpl) evt);
            // 当前节点任务实体

            // TODO 获取到了taskEntity 自己做每个节点的前置操作
            FlowableEventType eventType = event.getType();
            //待办创建
            if(eventType == FlowableEngineEventType.TASK_CREATED){
                TaskEntity taskEntity = (TaskEntity) event.getEntity();
                System.out.println("创建任务");
            }
            //待办分配
            else if(eventType == FlowableEngineEventType.TASK_ASSIGNED){
                TaskEntity taskEntity = (TaskEntity) event.getEntity();
                System.out.println("分配任务");
            }
            //待办完成
            else if(eventType == FlowableEngineEventType.TASK_COMPLETED){
                TaskEntity taskEntity = (TaskEntity) event.getEntity();
                System.out.println("完成任务");
            }
        }

    }

    @Override
    public boolean isFailOnException() {
        return false;
    }

    @Override
    public boolean isFireOnTransactionLifecycleEvent() {
        return false;
    }

    @Override
    public String getOnTransaction() {
        return null;
    }
}